/**@file 

// unDeviceTypeCreatorDeviceFunctions.ctl
This library contains functions used to manipulate device types in the package creation application.

@par Creation Date
  30/11/2011

@par Modification History
  30/11/2011 - First release
  
@par Constraints
  Can only be used by the main panel of the package creation application (dependencies to variables and widgets of the panel)

@par Usage
  Public

@par PVSS managers
  Ui, CTRL

@author 
  Cyril Caillaba (EN-ICE)
*/

//@{

// ------------------------------------------------------------------------------------ //
// ------------------unDeviceTypeCreatorDevice_addNewDeviceType--------------------- //
// ------------------------------------------------------------------------------------ //
/** Adds a new device type in the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL

@param sDeviceTypeName  input, name of the device type to add to the list.
@param bAlreadyExists input, whether the device already exists somewhere else or not.
@param bBefore intput, true to insert before the currently selected device, false to insert after.
@param dsExceptions output, the list of exceptions that occur in the function.
*/
void unDeviceTypeCreatorDevice_addNewDeviceType(
    string sDeviceTypeName,
    bool bAlreadyExists,
    bool bBefore,
    dyn_string &dsExceptions)
{
  
  // Check if the name is empty
  if (sDeviceTypeName == "")
  {
    string sError = getCatStr("unDeviceTypeCreator","EDIT-EMPTYDEVICENAME");
      
    fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "EDIT-EMPTYDEVICENAME"
    ); 
    return; 
  }
  
  // Always replace spaces by underscores
  strreplace(sDeviceTypeName, " ", "_");
  
    
  shape tableDevice = getShape("tableDeviceList"); 
  
  // Check if this device type is not already in the list
  for (int i = 0 ; i < tableDevice.lineCount() ; i++)
  {
    string sCurrentDeviceTypeName;
    getValue("tableDeviceList","cellValueRC",i,COLUMN_DEVICE_NAME,sCurrentDeviceTypeName);  
    if (sCurrentDeviceTypeName == sDeviceTypeName)
    {
      string sError = getCatStr("unDeviceTypeCreator","EDIT-DEVICEALREADYEXISTS");
      strreplace(sError,"DEVICE",sDeviceTypeName);
      
      fwException_raise(
        dsExceptions,
        "ERROR",
        sError,
        "EDIT-DEVICEALREADYEXISTS"
      ); 
      return; 
    }
      
  }    
 
  if (currentCellDevice > -1)
  {
    if (bBefore)
    {
      tableDevice.insertLineN(currentCellDevice);
      tableDevice.cellValueRC(currentCellDevice,COLUMN_DEVICE_NAME,sDeviceTypeName);
      tableDevice.cellValueRC(currentCellDevice,COLUMN_DEVICE_EXISTS,bAlreadyExists);
    }
    else
    {
      tableDevice.insertLineN(currentCellDevice+1);
      tableDevice.cellValueRC(currentCellDevice+1,COLUMN_DEVICE_NAME,sDeviceTypeName);
      tableDevice.cellValueRC(currentCellDevice+1,COLUMN_DEVICE_EXISTS,bAlreadyExists);
    }
    
  }
  else
  {
    if (bBefore)
    {
      tableDevice.insertLineN(0);
      tableDevice.cellValueRC(0,COLUMN_DEVICE_NAME,sDeviceTypeName);
      tableDevice.cellValueRC(0,COLUMN_DEVICE_EXISTS,bAlreadyExists);
    }
    else 
    {
      tableDevice.appendLine(COLUMN_DEVICE_NAME, sDeviceTypeName, COLUMN_DEVICE_EXISTS, bAlreadyExists);
    }
  }

  // Clear the selected line, so that we add at the end for the next device
  currentCellDevice = -1;
  
  // Clear the content of the text field
  textFieldNewDeviceType.text("");
}

// ------------------------------------------------------------------------------------ //
// -------------------unDeviceTypeCreatorDevice_deleteDeviceType-------------------- //
// ------------------------------------------------------------------------------------ //

/** Deletes the currently selected device type in the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL
*/
void unDeviceTypeCreatorDevice_deleteDeviceType()
{
  if (currentCellDevice == -1)
  {
    return;
  } 
  else
  {
    setValue("tableDeviceList","deleteLineN",currentCellDevice);
	// After the remove, clear the selection.
    currentCellDevice = -1;
  }
}

// ------------------------------------------------------------------------------------ //
// ----------------unDeviceTypeCreatorDevice_deleteAllDeviceTypes------------------- //
// ------------------------------------------------------------------------------------ //

/** Deletes all the device types in the list.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL
*/
void unDeviceTypeCreatorDevice_deleteAllDeviceTypes()
{
  int lineCount;
  getValue("tableDeviceList","lineCount",lineCount);
  setValue("tableDeviceList","deleteLinesN",0,lineCount);
  currentCellDevice = -1;
}

// ------------------------------------------------------------------------------------ //
// ------------------unDeviceTypeCreatorDevice_deviceExists------------------------- //
// ------------------------------------------------------------------------------------ //
/** Checks if a device exists. Exists means the matching checkbox is checked.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL

@param sDeviceName  input, name of the device type to check existance.
@param bExists output, whether the device is checked as existing or not.
@return -1 if the device type is not found in the list, 0 if everything goes well.
*/
int unDeviceTypeCreatorDevice_deviceExists(string sDeviceName, bool &bExists)
{
  bExists = false;
  for (int i = 1 ; i <= dynlen(deviceTypeList) ; i++)
  {
    if (deviceTypeList[i][LIST_NAME_INDEX] == sDeviceName)
    {
      bExists = deviceTypeList[i][LIST_EXISTS_INDEX];
      return 0;    
    }
  }
  
  return -1;
}

// ------------------------------------------------------------------------------------ //
// ------------------unDeviceTypeCreatorDevice_createNewDevice---------------------- //
// ------------------------------------------------------------------------------------ //

/** Creates all necessary information for a new device.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  Ui, CTRL

@param sDeviceType  input, the name of the device type to create.
@param mFileFunctions output, each key correspond to a file that will need to be created, and the value is a list of elements inside this file (functions mainly)
@param sTemplateSourcePath  input, the path where the template is located.
@param sPrefix  input, the prefix used for the generation.
@param sPackage  input, the name of the package.
@param sNewPackagePath  input, the path where to save the new package data.
@return -1 in case of failure, 0 else.
*/
int unDeviceTypeCreatorDevice_createNewDevice(string sDeviceType, mapping &mFileFunctions, string sTemplateSourcePath, string sPrefix, string sPackage, string sNewPackagePath)
{
       
  // ------------------------------------------------------------------ STEP 1 : DPT creation ------------------------------------------------------------------------//
      
  
  // Create the DPT also means exporting it to ASCII to be imported later.
  // So let's just generate an ASCII dump directly  
  // -------------------------------------- //
  // ----- File dplist/TAG_package.dpl ---- //
  // -------------------------------------- //   
  
  if (!mappingHasKey(mFileFunctions,"dplist/" + sPackage + ".dpl"))
  {
    mFileFunctions["dplist/" + sPackage + ".dpl"] = makeDynString();
    dynAppend(mFileFunctions["dplist/" + sPackage + ".dpl"],"# ascii dump of database\n\n");  
  }
  
  
  string dpTypeDevice = 
      "\n\n# DpType\n"
      "TypeName\n" +
      sDeviceType + "." + sDeviceType + "\t" + DPEL_STRUCT + "#1\n"
        "\tstatusInformation\t"   + DPEL_TYPEREF    + "#2:_UnStatusInformation\n"
        "\tProcessInput\t"        + DPEL_STRUCT     + "#3\n"
          "\t\tdata\t"            + DPEL_BLOB       + "#4\n"
          "\t\tiData\t"           + DPEL_INT        + "#14\n"
        "\tProcessOutput\t"       + DPEL_STRUCT     + "#24\n"
          "\t\tset\t"             + DPEL_INT        + "#25\n"      
      ;
   
  dynAppend(mFileFunctions["dplist/" + sPackage + ".dpl"],dpTypeDevice);  
  
  
  
      
  // ------------------------------------------------------------------ STEP 2 : device definition ------------------------------------------------------------------ //
  
  // Generation of ASCII dump
  unsigned newDpId;
  int dpeId;
  // Creation of a DP of type _UnObjects
  // If it already exists the old one is removed
  if (dpExists(sDeviceType+"_unicosInfo"))
  {  
    dpDelete(sDeviceType+"_unicosInfo");
  }
  dpCreate(sDeviceType+"_unicosInfo","_UnObjects");  
  dpGetId(sDeviceType+"_unicosInfo", newDpId, dpeId);
  
  string dpDevice = 
      "\n\n# Datapoint/DpId\n"
      "DpName	TypeName	ID\n" +
      sDeviceType+"_unicosInfo\t_UnObjects\t" + newDpId + "\n";
    
  // Creation of a DP of type _UnTrendConfiguration
  // If it already exists the old one is removed
  if (dpExists(sDeviceType+"_trendConfiguration"))
  {  
    dpDelete(sDeviceType+"_trendConfiguration");
  }
  dpCreate(sDeviceType+"_trendConfiguration","_UnTrendConfiguration");  
  dpGetId(sDeviceType+"_trendConfiguration", newDpId, dpeId);  
  
  dpDevice += 
      sDeviceType+"_trendConfiguration\t_UnTrendConfiguration\t" + newDpId + "\n";
  
  // Creation of a DP of type FwTrendingPlot
  // If it already exists the old one is removed
  if (dpExists("_template_un" + sDeviceType + "_Trend"))
  {  
    dpDelete("_template_un" + sDeviceType + "_Trend");
  }
  dpCreate("_template_un" + sDeviceType + "_Trend","FwTrendingPlot");  
  dpGetId("_template_un" + sDeviceType + "_Trend", newDpId, dpeId);  
  
  dpDevice += 
      "_template_un" + sDeviceType + "_Trend\tFwTrendingPlot\t" + newDpId + "\n";
  
  
  
  dynAppend(mFileFunctions["dplist/" + sPackage + ".dpl"],dpDevice);  
  
  
  // Generating dp values dump    
  string stringFileTAG_package_dpl;
  fileToString(sTemplateSourcePath + "/dplist/TAG_package.dpl",stringFileTAG_package_dpl);

  // Extract dpValues section
  string dpValuesSection = substr(stringFileTAG_package_dpl,strpos(stringFileTAG_package_dpl,"# DpValue") + 10); // +10 to skip the comment
  
  // Separate each line
  dyn_string dpValuesLines = strsplit(dpValuesSection,"\n");
  string dpValues = "\n\n# DpValue\n" + dpValuesLines[1] +"\n";
  
  for (int i = 2 ; i <= dynlen(dpValuesLines) ; i++)
  {
    string currentLine = dpValuesLines[i];
    if (strpos(currentLine,"TAG_DeviceType") >= 0)
    {
      // The line concerns a device, it is treated
      // TAG_* replacements
      
      
      // Date replacement
      // Date is at the end of the line, e.g. 30.11.2009 11:51:22.399 (23 last characters)
      currentLine = substr(currentLine, 0, strlen(currentLine) - 23);
      time now = getCurrentTime();
      currentLine += formatTime("%d.%m.%Y %H:%M:%S.",now,"%03d");
      strreplace(currentLine,"TAG_DeviceType",sDeviceType);
      strreplace(currentLine,"TAG_prefix",sPrefix);
      strreplace(currentLine,"TAG_package",sPackage);
      strreplace(currentLine,"TAG_PACKAGE",strtoupper(sPackage));
      
      // One of the lines contains a list of widgets which need to be filled
      // Note : could have been much simpler, but like this it will be easier to maintain in case of template change
      if (strpos(currentLine,"TAG_WidgetType") > 0)
      {
        // Iterate through the widgets attached to this device
        string currentLineBeginning = substr(currentLine,0,strpos(currentLine,"\"")); // The list begins at the first "
        string currentLineEndInclList = substr(currentLine,strpos(currentLine,"\"") + 1); // Second part of the string, including the list
        string singleElementInList = substr(currentLineEndInclList,0,strpos(currentLineEndInclList,"\"")); // Contains the single element in the list, without the 2 "
        string currentLineEnd = substr(currentLineEndInclList,strpos(currentLineEndInclList,"\"") + 1); // Second part of the string, without the list
        
        string widgetList;
        for (int j = 1 ; j <= dynlen(deviceWidgetMapping[sDeviceType]) ; j++)
        {
          string tempListElement = singleElementInList;
          strreplace(tempListElement,"TAG_WidgetType",deviceWidgetMapping[sDeviceType][j]);
          widgetList += "\"" + tempListElement + "\"";
          // Do not add the separation coma for the last element
          if ( j != dynlen(deviceWidgetMapping[sDeviceType]))
          {
            widgetList += ",";
          }
          
        }
        
        currentLine = currentLineBeginning + widgetList + currentLineEnd;
      }
      dpValues += currentLine + "\n";
    }
    // Else it is ignored
    
  }
  
  
  dynAppend(mFileFunctions["dplist/" + sPackage + ".dpl"],dpValues);  
  
  
  
  
  
  
  
  

  
  
  
  
  // Creation of a DP of type _FwDeviceDefinition
  // If it already exists the old one is removed
  if (dpExists(sDeviceType+"Info"))
  {  
    dpDelete(sDeviceType+"Info");
  }
  dpCreate(sDeviceType+"Info","_FwDeviceDefinition");  
  dpGetId(sDeviceType+"Info", newDpId, dpeId);
  
  
  if (!mappingHasKey(mFileFunctions,"dplist/" + sPackage + "fwDeviceDefinition.dpl"))
  {
    mFileFunctions["dplist/" + sPackage + "fwDeviceDefinition.dpl"] = makeDynString();
    dynAppend(mFileFunctions["dplist/" + sPackage + "fwDeviceDefinition.dpl"],"# ascii dump of database\n\n");  
  }
  
  dynAppend(mFileFunctions["dplist/" + sPackage + "fwDeviceDefinition.dpl"],
      "\n# Datapoint/DpId\n"
      "DpName	TypeName	ID\n" + 
      sDeviceType + "Info\t_FwDeviceDefinition\t" + newDpId + "\n\n");  
  
  string stringFileTAG_packagefwDeviceDefinition_dpl;
  fileToString(sTemplateSourcePath + "/dplist/TAG_packagefwDeviceDefinition.dpl",stringFileTAG_packagefwDeviceDefinition_dpl);
  dpValues = substr(stringFileTAG_packagefwDeviceDefinition_dpl,strpos(stringFileTAG_packagefwDeviceDefinition_dpl,"# DpValue")); 
  dpValuesLines = strsplit(dpValues,"\n");
  
  
  dynAppend(mFileFunctions["dplist/" + sPackage + "fwDeviceDefinition.dpl"],dpValuesLines[1] + "\n");
  dynAppend(mFileFunctions["dplist/" + sPackage + "fwDeviceDefinition.dpl"],dpValuesLines[2] + "\n");
  // Start at i = 3 to exclude "# DpValue" and "ElementName	TypeName	_original.._value	_original.._status	_original.._stime"
  for (int i = 3 ; i <= dynlen(dpValuesLines) ; i++)
  {
    string currentLine = dpValuesLines[i];
      
    // Date replacement
    // Date is at the end of the line, e.g. 30.11.2009 11:51:22.399 (23 last characters)
    currentLine = substr(currentLine, 0, strlen(currentLine) - 23);
    time now = getCurrentTime();
    currentLine += formatTime("%d.%m.%Y %H:%M:%S.",now,"%03d");
    strreplace(currentLine,"TAG_DeviceType",sDeviceType);
    dynAppend(mFileFunctions["dplist/" + sPackage + "fwDeviceDefinition.dpl"],currentLine + "\n");
  }    
  
  
  
  
  // ------------------------------------------------------------------ STEP 3 : JCOP device definition ------------------------------------------------------------------ //
  // Nothing to do in the document
  // ------------------------------------------------------------------ STEP 4 : create device check & import functions ------------------------------------------------------------------ //  

  // List of the front-end types using this device
  dyn_string mappedFrontEndTypes = makeDynString();
  for (int i = 1 ; i <= dynlen(mappingKeys(frontEndDeviceMapping)) ; i++)
  {
    string frontEndType = mappingKeys(frontEndDeviceMapping)[i];
    if (dynContains(frontEndDeviceMapping[frontEndType],sDeviceType) > 0)
    {
      dynAppend(mappedFrontEndTypes,frontEndType);
    } 
  }

  
  // --------------------------------------------------------------- //
  // ----- File scripts\libs\TAG_prefixConfigTAG_DeviceType.ctl ---- //
  // --------------------------------------------------------------- //       
  
  string stringFileTAG_prefixConfigTAG_DeviceType;
  fileToString(sTemplateSourcePath + "/scripts/libs/TAG_prefixConfigTAG_DeviceType.ctl",stringFileTAG_prefixConfigTAG_DeviceType);

  // Mapping initialisation 
  if (!mappingHasKey(mFileFunctions,"scripts/libs/" + sPrefix + "Config" + sDeviceType + ".ctl"))
  {
    mFileFunctions["scripts/libs/" + sPrefix + "Config" + sDeviceType + ".ctl"] = makeDynString();
      
    // Copy the file header
    string header = substr(stringFileTAG_prefixConfigTAG_DeviceType,0,strpos(stringFileTAG_prefixConfigTAG_DeviceType,"//@{")) + "//@{\n\n";   
    strreplace(header,"TAG_DeviceType",sDeviceType); 
    strreplace(header,"TAG_prefix",sPrefix);
    dynAppend(mFileFunctions["scripts/libs/" + sPrefix + "Config" + sDeviceType + ".ctl"],header);    
  }
    
  
  dyn_string functionListTAG_prefixConfigTAG_DeviceType = makeDynString();
  dynAppend(functionListTAG_prefixConfigTAG_DeviceType,"TAG_prefixConfigTAG_DeviceType_checkConfig");
  dynAppend(functionListTAG_prefixConfigTAG_DeviceType,"TAG_prefixConfigTAG_DeviceType_setConfig");
  dynAppend(functionListTAG_prefixConfigTAG_DeviceType,"TAG_FrontEndType_TAG_DeviceType_checkConfig");
  dynAppend(functionListTAG_prefixConfigTAG_DeviceType,"TAG_FrontEndType_TAG_DeviceType_setConfig");
  dynAppend(functionListTAG_prefixConfigTAG_DeviceType,"TAG_DeviceType_getArchiveDPE");
  dynAppend(functionListTAG_prefixConfigTAG_DeviceType,"TAG_DeviceType_getArchiveProxyDPE");
  

  
  
  
  
  for (int i = 1 ; i <= dynlen(functionListTAG_prefixConfigTAG_DeviceType) ; i++)
  {
    string functionContent = extractFunctionFromStringFile(stringFileTAG_prefixConfigTAG_DeviceType,functionListTAG_prefixConfigTAG_DeviceType[i]);
    // Tag replacement
    strreplace(functionContent,"TAG_prefix",sPrefix); 
    strreplace(functionContent,"TAG_DeviceType",sDeviceType); 
    strreplace(functionContent,"TAG_DEVICETYPE",strtoupper(sDeviceType));
    strreplace(functionContent,"TAG_PREFIX",strtoupper(sPrefix));
    
    // Some functions need to be repeated for each front-end linked to this device
    if (strpos(functionContent,"TAG_FrontEndType") > 0)
    {
      for (int j = 1 ; j <= dynlen(mappedFrontEndTypes) ; j++)
      {
        string tempFunctionContent = functionContent;
        strreplace(tempFunctionContent,"TAG_FrontEndType",mappedFrontEndTypes[j]);    
        dynAppend(mFileFunctions["scripts/libs/" + sPrefix + "Config" + sDeviceType + ".ctl"],tempFunctionContent);       
      }
    }
    else
    {
      dynAppend(mFileFunctions["scripts/libs/" + sPrefix + "Config" + sDeviceType + ".ctl"],functionContent);  
    }
  }
    
  
    
  
    
  // --------------------------------------------------------- //
  // ---- File scripts/libs/TAG_prefixTAG_DeviceType.ctl ----- //
  // --------------------------------------------------------- //   
  
  string stringFileTAG_prefixTAG_DeviceType;
  fileToString(sTemplateSourcePath + "/scripts/libs/TAG_prefixTAG_DeviceType.ctl",stringFileTAG_prefixTAG_DeviceType);

    
  if (!mappingHasKey(mFileFunctions,"scripts/libs/" + sPrefix + sDeviceType + ".ctl"))
  {
    // Mapping initialisation 
    mFileFunctions["scripts/libs/" + sPrefix + sDeviceType + ".ctl"] = makeDynString();
    
    // Copy the file header (only the first time, because every device is saved in this file)
    string header = substr(stringFileTAG_prefixTAG_DeviceType,0,strpos(stringFileTAG_prefixTAG_DeviceType,"//@{")) + "//@{\n\n";   
    strreplace(header,"TAG_prefix",sPrefix);
    strreplace(header,"TAG_DeviceType",sDeviceType);
    dynAppend(mFileFunctions["scripts/libs/" + sPrefix + sDeviceType + ".ctl"],header);
  }

    
  
  // Replacement is proceeded each function at a time to have a better control of the process and possible future changes
  dyn_string functionListTAG_prefixTAG_DeviceType = makeDynString();
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_AcknowledgeAlarm");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_ObjectListGetValueTime");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_FaceplateStatusRegisterCB");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_FaceplateStatusDisconnection");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_FaceplateConnect");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_FaceplateDisconnect");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_FaceplateStatusAnimationCB");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_ButtonRegisterCB");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_UserLoginGetButtonState");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_ButtonSetState");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_ButtonDisconnection");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_ButtonUserAccess");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_ButtonAnimationCB");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_WidgetRegisterCB");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_WidgetDisconnection");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_WidgetConnect");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_WidgetDisconnect");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_WidgetCB");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_MenuConfiguration");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_HandleMenu");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_TAG_WidgetTypeAnimation");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_TAG_WidgetTypeDisconnection");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_WidgetRegisterAlertCB");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_WidgetLockCallBack");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_WidgetConnectAlert");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_WidgetDisconnectAlert");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_FaceplateStatusRegisterAlertCB");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_FaceplateConnectAlert");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_FaceplateDisconnectAlert");
  dynAppend(functionListTAG_prefixTAG_DeviceType,"TAG_prefixTAG_DeviceType_FaceplateLockCallBack");
  
  
  for (int i = 1 ; i <= dynlen(functionListTAG_prefixTAG_DeviceType) ; i++)
  {
    string functionContent = extractFunctionFromStringFile(stringFileTAG_prefixTAG_DeviceType,functionListTAG_prefixTAG_DeviceType[i]);
    // Tag replacement
    strreplace(functionContent,"TAG_prefix",sPrefix); 
    strreplace(functionContent,"TAG_DeviceType",sDeviceType); 
    strreplace(functionContent,"TAG_DEVICETYPE",strtoupper(sDeviceType));
    strreplace(functionContent,"TAG_PREFIX",strtoupper(sPrefix));   
    
    
     // Some functions need to be repeated for each widget linked to this device
    if (strpos(functionContent,"TAG_WidgetType") > 0)
    {
      for (int j = 1 ; j <= dynlen(deviceWidgetMapping[sDeviceType]) ; j++)
      {
        string tempFunctionContent = functionContent;
        strreplace(tempFunctionContent,"TAG_WidgetType",deviceWidgetMapping[sDeviceType][j]);    
      dynAppend(mFileFunctions["scripts/libs/" + sPrefix + sDeviceType + ".ctl"],tempFunctionContent);       
      }
    }
    else
    {
      dynAppend(mFileFunctions["scripts/libs/" + sPrefix + sDeviceType + ".ctl"],functionContent);   
    }
  }

  

  // ------------------------------------------------------------------ STEP 5 : create export functions ------------------------------------------------------------------ //    
  
  
  
          
  // --------------------------------------------------------- //
  // ----- File scripts/libs/TAG_prefixExportDevices.ctl ----- //
  // --------------------------------------------------------- //     
      
  string stringFileTAG_prefixExportDevices;
  fileToString(sTemplateSourcePath + "/scripts/libs/TAG_prefixExportDevices.ctl",stringFileTAG_prefixExportDevices);

    
  if (!mappingHasKey(mFileFunctions,"scripts/libs/" + sPrefix + "ExportDevices.ctl"))
  {
    // Mapping initialisation 
    mFileFunctions["scripts/libs/" + sPrefix + "ExportDevices.ctl"] = makeDynString();
    
    // Copy the file header (only the first time, because this file can be accessed multiple times during package generation)
    string header = substr(stringFileTAG_prefixExportDevices,0,strpos(stringFileTAG_prefixExportDevices,"//@{")) + "//@{\n\n";   
    strreplace(header,"TAG_prefix",sPrefix);
    dynAppend(mFileFunctions["scripts/libs/" + sPrefix + "ExportDevices.ctl"],header);
  }

   // Extract desired function from the file
  string functionTAG_FrontEndType_TAG_DeviceType_ExportConfig = extractFunctionFromStringFile(stringFileTAG_prefixExportDevices,"TAG_FrontEndType_TAG_DeviceType_ExportConfig");

  // This function appears once for each front-end type using this device
  for (int i = 1 ; i <= dynlen(mappedFrontEndTypes) ; i++)
  {
    string frontEndType = mappedFrontEndTypes[i];
      
    // Temporary copy of the string not to erase the tags
    string tempFunctionTAG_FrontEndType_TAG_DeviceType_ExportConfig = functionTAG_FrontEndType_TAG_DeviceType_ExportConfig;
    
    strreplace(tempFunctionTAG_FrontEndType_TAG_DeviceType_ExportConfig,"TAG_prefix",sPrefix);
    strreplace(tempFunctionTAG_FrontEndType_TAG_DeviceType_ExportConfig,"TAG_PREFIX",strtoupper(sPrefix));
    strreplace(tempFunctionTAG_FrontEndType_TAG_DeviceType_ExportConfig,"TAG_DeviceType",sDeviceType);
    strreplace(tempFunctionTAG_FrontEndType_TAG_DeviceType_ExportConfig,"TAG_FRONTENDTYPE",strtoupper(frontEndType));
    strreplace(tempFunctionTAG_FrontEndType_TAG_DeviceType_ExportConfig,"TAG_FrontEndType",frontEndType);
    dynAppend(mFileFunctions["scripts/libs/" + sPrefix + "ExportDevices.ctl"],tempFunctionTAG_FrontEndType_TAG_DeviceType_ExportConfig);
      
  }
      

  // ------------------------------------------------------------------ STEP 6 : create the device functions ------------------------------------------------------------------ //   
  // Done in step 4
    
  
  
  // ------------------------------------------------------------------ STEP 7 : create the device icons ------------------------------------------------------------ //   
  
  
  
  
  // Just copy the icons provided in the template folder
  copyFile(sTemplateSourcePath + "/pictures/TAG_DeviceType/iconWithChildrenClose.bmp",          sNewPackagePath + "/" + sPackage + "/pictures/" + sDeviceType + "/iconWithChildrenClose.bmp");
  int xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE); 
  xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./pictures/" + sDeviceType + "/iconWithChildrenClose.bmp");
  copyFile(sTemplateSourcePath + "/pictures/TAG_DeviceType/iconWithChildrenOpen.bmp",           sNewPackagePath + "/" + sPackage + "/pictures/" + sDeviceType + "/iconWithChildrenOpen.bmp");
  xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE); 
  xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./pictures/" + sDeviceType + "/iconWithChildrenOpen.bmp");
  copyFile(sTemplateSourcePath + "/pictures/TAG_DeviceType/iconWithChildrenPath.bmp",           sNewPackagePath + "/" + sPackage + "/pictures/" + sDeviceType + "/iconWithChildrenPath.bmp");
  xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE); 
  xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./pictures/" + sDeviceType + "/iconWithChildrenPath.bmp");
  copyFile(sTemplateSourcePath + "/pictures/TAG_DeviceType/iconWithChildrenSelected.bmp",       sNewPackagePath + "/" + sPackage + "/pictures/" + sDeviceType + "/iconWithChildrenSelected.bmp");
  xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE); 
  xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./pictures/" + sDeviceType + "/iconWithChildrenSelected.bmp");
  copyFile(sTemplateSourcePath + "/pictures/TAG_DeviceType/iconWithoutChildrenNotSelected.bmp", sNewPackagePath + "/" + sPackage + "/pictures/" + sDeviceType + "/iconWithoutChildrenNotSelected.bmp");
  xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE); 
  xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./pictures/" + sDeviceType + "/iconWithoutChildrenNotSelected.bmp");
  copyFile(sTemplateSourcePath + "/pictures/TAG_DeviceType/iconWithoutChildrenSelected.bmp",    sNewPackagePath + "/" + sPackage + "/pictures/" + sDeviceType + "/iconWithoutChildrenSelected.bmp");
  xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE); 
  xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./pictures/" + sDeviceType + "/iconWithoutChildrenSelected.bmp");
       
  // Necessary functions have been treated in step 4
  
  
  
  
  // ------------------------------------------------------------------ STEP 8 : create the widget, catalog of widget, widget param ------------------------------------------------------------------ //   
  
    
  // -------------------------------------------------------------------------------------------- //
  // ------ File panels/objects_parampanels/TAG_PACKAGE/TAG_prefixTAG_DeviceType_para.pnl ------- //
  // -------------------------------------------------------------------------------------------- //   
  

  string stringFileTAG_prefixTAG_DeviceType_para;
  fileToString(sTemplateSourcePath + "/panels/objects_parampanels/TAG_PACKAGE/TAG_prefixTAG_DeviceType_para.pnl",stringFileTAG_prefixTAG_DeviceType_para);

  mFileFunctions["panels/objects_parampanels/" + strtoupper(sPackage) + "/" + sPrefix + sDeviceType + "_para.pnl"] = makeDynString(); 
  strreplace(stringFileTAG_prefixTAG_DeviceType_para,"TAG_DeviceType",sDeviceType);
  dynAppend(mFileFunctions["panels/objects_parampanels/" + strtoupper(sPackage) + "/" + sPrefix + sDeviceType + "_para.pnl"],stringFileTAG_prefixTAG_DeviceType_para);
  
  
  
  // -------------------------------------------------------------------------------------------- //
  // ----- File panels/objects/TAG_PACKAGE/TAG_prefixWidgetTAG_DeviceTypeTAG_WidgetType.pnl ----- //
  // -------------------------------------------------------------------------------------------- //   

  string stringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetType;
  fileToString(sTemplateSourcePath + "/panels/objects/TAG_PACKAGE/TAG_prefixWidget_TAG_DeviceTypeTAG_WidgetType.pnl",stringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetType);
  
  // Iterate through the widgets of this device  
  // Note : this actions should be done only if the widget does not exist yet. But since the device does not exist, all its widget also do not exist.
  for (int i = 1 ; i <= dynlen(deviceWidgetMapping[sDeviceType]) ; i++)
  {
    string currentWidget = deviceWidgetMapping[sDeviceType][i];
    string tempStringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetType = stringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetType;
    if (!mappingHasKey(mFileFunctions,"panels/objects/" + strtoupper(sPackage) + "/" + sPrefix + "Widget_" + sDeviceType + currentWidget + ".pnl"))
    {
      mFileFunctions["panels/objects/" + strtoupper(sPackage) + "/" + sPrefix + "Widget_" + sDeviceType + currentWidget + ".pnl"] = makeDynString();    
    }
    // Warning, a line is "LANG:1 14 TAG_DeviceType", where 14 is the size of the following string. This needs to be fixed
    strreplace(tempStringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetType, "LANG:1 14 TAG_DeviceType", "LANG:1 "+strlen(sDeviceType)+" "+sDeviceType);
    strreplace(tempStringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetType, "TAG_PACKAGE", strtoupper(sPackage));
    strreplace(tempStringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetType, "TAG_prefix", sPrefix);
    strreplace(tempStringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetType, "TAG_DeviceType", sDeviceType);
    strreplace(tempStringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetType, "TAG_WidgetType", currentWidget);
    
    dynAppend(mFileFunctions["panels/objects/" + strtoupper(sPackage) + "/" + sPrefix + "Widget_" + sDeviceType + currentWidget + ".pnl"],tempStringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetType);
  
  }
  
  
  // ---------------------------------------------------------------------------------------------- //
  // - File panels/objects/TAG_PACKAGE/TAG_prefixWidget_TAG_DeviceTypeTAG_WidgetTypeWithAlert.pnl - //
  // ---------------------------------------------------------------------------------------------- //   

  file fileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetTypeWithAlert = fopen(sTemplateSourcePath + "/panels/objects/TAG_PACKAGE/TAG_prefixWidget_TAG_DeviceTypeTAG_WidgetTypeWithAlert.pnl", "r");
  string stringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetTypeWithAlert;
  fileToString(sTemplateSourcePath + "/panels/objects/TAG_PACKAGE/TAG_prefixWidget_TAG_DeviceTypeTAG_WidgetTypeWithAlert.pnl",stringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetTypeWithAlert);
  fclose(fileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetTypeWithAlert);
  
  // Iterate through the widgets of this device  
  for (int i = 1 ; i <= dynlen(deviceWidgetMapping[sDeviceType]) ; i++)
  {
    string currentWidget = deviceWidgetMapping[sDeviceType][i];
    string tempStringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetTypeWithAlert = stringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetTypeWithAlert;
    
    if (!mappingHasKey(mFileFunctions,"panels/objects/" + strtoupper(sPackage) + "/" + sPrefix + "Widget_" + sDeviceType + currentWidget + "WithAlert.pnl"))
    {
      mFileFunctions["panels/objects/" + strtoupper(sPackage) + "/" + sPrefix + "Widget_" + sDeviceType + currentWidget + "WithAlert.pnl"] = makeDynString();    
    }
    // Warning, a line is "LANG:1 14 TAG_DeviceType", where 14 is the size of the following string. This needs to be fixed
    strreplace(tempStringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetTypeWithAlert, "LANG:1 14 TAG_DeviceType", "LANG:1 "+strlen(sDeviceType)+" "+sDeviceType);
    strreplace(tempStringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetTypeWithAlert, "TAG_PACKAGE", strtoupper(sPackage));
    strreplace(tempStringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetTypeWithAlert, "TAG_prefix", sPrefix);
    strreplace(tempStringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetTypeWithAlert, "TAG_DeviceType", sDeviceType);
    strreplace(tempStringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetTypeWithAlert, "TAG_WidgetType", currentWidget);
    
    dynAppend(mFileFunctions["panels/objects/" + strtoupper(sPackage) + "/" + sPrefix + "Widget_" + sDeviceType + currentWidget + "WithAlert.pnl"],tempStringFileTAG_prefixWidgetTAG_DeviceTypeTAG_WidgetTypeWithAlert);
  
  }
  
  
  // Copy images  
  for (int i = 1 ; i <= dynlen(deviceWidgetMapping[sDeviceType]) ; i++)
  {
    string currentWidget = deviceWidgetMapping[sDeviceType][i];
    
  copyFile(
      sTemplateSourcePath + "/images/objects/TAG_PACKAGE/TAG_prefixWidget_TAG_DeviceTypeTAG_WidgetType.pnl/TAG_prefixWidget_TAG_DeviceTypeTAG_WidgetType.png",
      sNewPackagePath +     "/" + sPackage + "/" + "/images/objects/"+sPackage+"/"+sPrefix+"Widget_"+sDeviceType+currentWidget+".pnl/"+sPrefix+"Widget_"+sDeviceType+currentWidget+".png");
  copyFile(
      sTemplateSourcePath + "/images/objects/TAG_PACKAGE/TAG_prefixWidget_TAG_DeviceTypeTAG_WidgetTypeWithAlert.pnl/TAG_prefixWidget_TAG_DeviceTypeTAG_WidgetTypeWithAlert.png",
      sNewPackagePath +     "/" + sPackage + "/" + "/images/objects/"+sPackage+"/"+sPrefix+"Widget_"+sDeviceType+currentWidget+"WithAlert.pnl/"+sPrefix+"Widget_"+sDeviceType+currentWidget+"WithAlert.png");
  
    xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE);
    xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./images/objects/"+sPackage+"/"+sPrefix+"Widget_"+sDeviceType+currentWidget+".pnl/"+sPrefix+"Widget_"+sDeviceType+currentWidget+".png");
  
    xmlNodeIdFile = xmlAppendChild(packageComponentXML,xmlNodeIdComponent,XML_ELEMENT_NODE,XML_NODE_FILE);
    xmlAppendChild(packageComponentXML,xmlNodeIdFile,XML_TEXT_NODE,"./images/objects/"+sPackage+"/"+sPrefix+"Widget_"+sDeviceType+currentWidget+"WithAlert.pnl/"+sPrefix+"Widget_"+sDeviceType+currentWidget+"WithAlert.png");
  }
  
  
  
  // Necessary functions have been treated in step 4
  
  
  
  // ------------------------------------------------------------------ STEP 9 : faceplate ------------------------------------------------------------ //  


  // --------------------------------------------------------------------------------------------- //
  // ---- File panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixFaceplateTAG_DeviceType.pnl ---- //
  // --------------------------------------------------------------------------------------------- //  
  
      
  file fileTAG_prefixFaceplateTAG_DeviceType = fopen(sTemplateSourcePath + "/panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixFaceplateTAG_DeviceType.pnl", "r");
  string stringFileTAG_prefixFaceplateTAG_DeviceType;
  fileToString(sTemplateSourcePath + "/panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixFaceplateTAG_DeviceType.pnl",stringFileTAG_prefixFaceplateTAG_DeviceType);
  fclose(fileTAG_prefixFaceplateTAG_DeviceType);
  
  mFileFunctions["panels/vision/" + strtoupper(sPackage) + "/" + sDeviceType + "/" + sPrefix + "FacePlate" + sDeviceType + ".pnl"] = makeDynString();
  
  strreplace(stringFileTAG_prefixFaceplateTAG_DeviceType,"TAG_DeviceType",sDeviceType);
  strreplace(stringFileTAG_prefixFaceplateTAG_DeviceType,"TAG_package",sPackage);
  strreplace(stringFileTAG_prefixFaceplateTAG_DeviceType,"TAG_prefix",sPrefix);
  
  dynAppend(mFileFunctions["panels/vision/" + strtoupper(sPackage) + "/" + sDeviceType + "/" + sPrefix + "FacePlate" + sDeviceType + ".pnl"],stringFileTAG_prefixFaceplateTAG_DeviceType);  
  
  
  
  // ------------------------------------------------------------------------------------------------------ //
  // ---- File panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixFaceplateTAG_DeviceTypeWithAlert.pnl ---- //
  // ------------------------------------------------------------------------------------------------------ //  
  
      
  file fileTAG_prefixFaceplateTAG_DeviceTypeWithAlert = fopen(sTemplateSourcePath + "/panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixFaceplateTAG_DeviceTypeWithAlert.pnl", "r");
  string stringFileTAG_prefixFaceplateTAG_DeviceTypeWithAlert;
  fileToString(sTemplateSourcePath + "/panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixFaceplateTAG_DeviceTypeWithAlert.pnl",stringFileTAG_prefixFaceplateTAG_DeviceTypeWithAlert);
  fclose(fileTAG_prefixFaceplateTAG_DeviceTypeWithAlert);
  
  mFileFunctions["panels/vision/" + strtoupper(sPackage) + "/" + sDeviceType + "/" + sPrefix + "FacePlate" + sDeviceType + "WithAlert.pnl"] = makeDynString();
  
  strreplace(stringFileTAG_prefixFaceplateTAG_DeviceTypeWithAlert,"TAG_DeviceType",sDeviceType);
  strreplace(stringFileTAG_prefixFaceplateTAG_DeviceTypeWithAlert,"TAG_package",sPackage);
  strreplace(stringFileTAG_prefixFaceplateTAG_DeviceTypeWithAlert,"TAG_prefix",sPrefix);
  
  dynAppend(mFileFunctions["panels/vision/" + strtoupper(sPackage) + "/" + sDeviceType + "/" + sPrefix + "FacePlate" + sDeviceType + "WithAlert.pnl"],stringFileTAG_prefixFaceplateTAG_DeviceTypeWithAlert); 
  
  // ---------------------------------------------------------------------------------------------------- //
  // ---- File panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixFaceplateTAG_DeviceType_status.pnl ---- //
  // ---------------------------------------------------------------------------------------------------- //  
  
      
  file fileTAG_prefixFaceplateTAG_DeviceType_status = fopen(sTemplateSourcePath + "/panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixFaceplateTAG_DeviceType_status.pnl", "r");
  string stringFileTAG_prefixFaceplateTAG_DeviceType_status;
  fileToString(sTemplateSourcePath + "/panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixFaceplateTAG_DeviceType_status.pnl",stringFileTAG_prefixFaceplateTAG_DeviceType_status);
  fclose(fileTAG_prefixFaceplateTAG_DeviceType_status);
  
  mFileFunctions["panels/vision/" + strtoupper(sPackage) + "/" + sDeviceType + "/" + sPrefix + "FacePlate" + sDeviceType + "_status.pnl"] = makeDynString();
  
  strreplace(stringFileTAG_prefixFaceplateTAG_DeviceType_status,"TAG_DeviceType",sDeviceType);
  strreplace(stringFileTAG_prefixFaceplateTAG_DeviceType_status,"TAG_package",sPackage);
  strreplace(stringFileTAG_prefixFaceplateTAG_DeviceType_status,"TAG_prefix",sPrefix);
  
  dynAppend(mFileFunctions["panels/vision/" + strtoupper(sPackage) + "/" + sDeviceType + "/" + sPrefix + "FacePlate" + sDeviceType + "_status.pnl"],stringFileTAG_prefixFaceplateTAG_DeviceType_status); 
  
  // ------------------------------------------------------------------------------------------------------------- //
  // ---- File panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixFaceplateTAG_DeviceType_statusWithAlert.pnl ---- //
  // ------------------------------------------------------------------------------------------------------------- //   
  
      
  file fileTAG_prefixFaceplateTAG_DeviceType_statusWithAlert = fopen(sTemplateSourcePath + "/panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixFaceplateTAG_DeviceType_statusWithAlert.pnl", "r");
  string stringFileTAG_prefixFaceplateTAG_DeviceType_statusWithAlert;
  fileToString(sTemplateSourcePath + "/panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixFaceplateTAG_DeviceType_statusWithAlert.pnl",stringFileTAG_prefixFaceplateTAG_DeviceType_statusWithAlert);
  fclose(fileTAG_prefixFaceplateTAG_DeviceType_statusWithAlert);
  
  mFileFunctions["panels/vision/" + strtoupper(sPackage) + "/" + sDeviceType + "/" + sPrefix + "FacePlate" + sDeviceType + "_statusWithAlert.pnl"] = makeDynString();
  
  strreplace(stringFileTAG_prefixFaceplateTAG_DeviceType_statusWithAlert,"TAG_DeviceType",sDeviceType);
  strreplace(stringFileTAG_prefixFaceplateTAG_DeviceType_statusWithAlert,"TAG_package",sPackage);
  strreplace(stringFileTAG_prefixFaceplateTAG_DeviceType_statusWithAlert,"TAG_prefix",sPrefix);
  
  dynAppend(mFileFunctions["panels/vision/" + strtoupper(sPackage) + "/" + sDeviceType + "/" + sPrefix + "FacePlate" + sDeviceType + "_statusWithAlert.pnl"],stringFileTAG_prefixFaceplateTAG_DeviceType_statusWithAlert);  
  


  // ------------------------------------------------------------------ STEP 10 : trending ------------------------------------------------------------ //  

  // ----------------------------------------------------------------------------------------------------- //
  // ---- File panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixTAG_DeviceTypeTrendingFaceplate.pnl ---- //
  // ----------------------------------------------------------------------------------------------------- //   
  
      
  file fileTAG_prefixTAG_DeviceTypeTrendingFaceplate = fopen(sTemplateSourcePath + "/panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixTAG_DeviceTypeTrendingFaceplate.pnl", "r");
  string stringFileTAG_prefixTAG_DeviceTypeTrendingFaceplate;
  fileToString(sTemplateSourcePath + "/panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixTAG_DeviceTypeTrendingFaceplate.pnl",stringFileTAG_prefixTAG_DeviceTypeTrendingFaceplate);
  fclose(fileTAG_prefixTAG_DeviceTypeTrendingFaceplate);
  
  mFileFunctions["panels/vision/" + strtoupper(sPackage) + "/" + sDeviceType + "/" + sPrefix + sDeviceType + "TrendingFaceplate.pnl"] = makeDynString();
  
  strreplace(stringFileTAG_prefixTAG_DeviceTypeTrendingFaceplate,"TAG_DeviceType",sDeviceType);
  strreplace(stringFileTAG_prefixTAG_DeviceTypeTrendingFaceplate,"TAG_package",sPackage);
  strreplace(stringFileTAG_prefixTAG_DeviceTypeTrendingFaceplate,"TAG_prefix",sPrefix);
  
  dynAppend(mFileFunctions["panels/vision/" + strtoupper(sPackage) + "/" + sDeviceType + "/" + sPrefix + sDeviceType + "TrendingFaceplate.pnl"],stringFileTAG_prefixTAG_DeviceTypeTrendingFaceplate);  


  // ------------------------------------------------------------------ STEP 11 : action interface ------------------------------------------------------------ //  
  
  // Necessary functions have been treated in step 4  
  
  // ------------------------------------------------------------------------------------------------ //
  // ---- File panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixTAG_DeviceType_ButtonPanel.pnl ---- //
  // ------------------------------------------------------------------------------------------------ //  
  
      
  file fileTAG_prefixTAG_prefixTAG_DeviceType_ButtonPanel = fopen(sTemplateSourcePath + "/panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixTAG_DeviceType_ButtonPanel.pnl", "r");
  string stringFileTAG_prefixTAG_DeviceType_ButtonPanel;
  fileToString(sTemplateSourcePath + "/panels/vision/TAG_PACKAGE/TAG_DeviceType/TAG_prefixTAG_DeviceType_ButtonPanel.pnl",stringFileTAG_prefixTAG_DeviceType_ButtonPanel);
  fclose(fileTAG_prefixTAG_prefixTAG_DeviceType_ButtonPanel);
  
  mFileFunctions["panels/vision/" + strtoupper(sPackage) + "/" + sDeviceType + "/" + sPrefix + sDeviceType + "_ButtonPanel.pnl"] = makeDynString();
  
  strreplace(stringFileTAG_prefixTAG_DeviceType_ButtonPanel,"TAG_DeviceType",sDeviceType);
  strreplace(stringFileTAG_prefixTAG_DeviceType_ButtonPanel,"TAG_package",sPackage);
  strreplace(stringFileTAG_prefixTAG_DeviceType_ButtonPanel,"TAG_prefix",sPrefix);  
  dynAppend(mFileFunctions["panels/vision/" + strtoupper(sPackage) + "/" + sDeviceType + "/" + sPrefix + sDeviceType + "_ButtonPanel.pnl"],stringFileTAG_prefixTAG_DeviceType_ButtonPanel);          
  

  
  
  
  
  
  // The device now exists, the appropriate flag must be changed so that it is not created twice
  
  // Changing in the device list used for generation
  int devicePos = dynContains(getDynString(deviceTypeList,LIST_NAME_INDEX),sDeviceType);
  if (devicePos > 0)
  {
    deviceTypeList[devicePos][LIST_EXISTS_INDEX] = true;
  }
  
  // Changing in the table  
  for (int i = 0 ; i < tableDeviceList.lineCount ; i++)
  {
    string currentDeviceName;
    getValue("tableDeviceList","cellValueRC",i,COLUMN_DEVICE_NAME,currentDeviceName);  
    if (currentDeviceName == sDeviceType)
    {
      setValue("tableDeviceList","cellValueRC",i,COLUMN_DEVICE_EXISTS,true);  
    }
  }
  
}


//@}
